/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
using namespace std;


#if !__SYNTHESIS__
    #define DBG 1
    #define DBG_INPUT 0
    #define DBG_ROUND 0
#else
    #define DBG 0
#endif

#define LANE_SIZE 16
typedef ap_uint<LANE_SIZE> tKeccakLane;

enum fsms_t {
    S_START, S_DATA,
    S_STRIDE, S_TAG };

#define KeccakF_width 400
#define maxRound 19
#define nrLanes 25
#define Ket_StartRoundOffset 8 
#define Ket_StrideRoundOffset 14 

#define FRAMEBITSEMPTY  0x01
#define FRAMEBITS0      0x02
#define FRAMEBITS00     0x04
#define FRAMEBITS10     0x05
#define FRAMEBITS01     0x06
#define FRAMEBITS11     0x07

#define index(x, y) (((x)%5)+5*((y)%5))
#define index2(x)   ((x%5)*5+

    
/* ------------------------------------------------------------------------- */
void ppp(const char* str, const tKeccakLane *s)
{
#if DBG
    printf("%s", str);
    ap_uint<KeccakF_width> state = 0;


    for(size_t i=0; i<nrLanes; i++){
      printf("%02X", (int)(s[i] & 0xFF));
      printf("%02X", (int)(s[i] >> 8));
    }

    printf("\n");
#endif
}

void pp(const char* str)
{
#if DBG
    printf("%s", str);
#endif
}

void printHeader(fsms_t fsm, int rnd, int ctr, int frame, 
                 public_bus bdi, int is_valid, int is_write)
{
#if DBG
    switch(fsm) {
    case S_START:   printf("S_START  : "); break;
    case S_DATA:    printf("S_DATA : "); break;
    case S_STRIDE:  printf("S_STRIDE : "); break;
    default:        printf("S_TAG    : ");
    }
    printf(" rnd[%2d] ctr[%d] frame[%02X] eot(%d) eoi(%d) valid(%d) write(%d)-> \n", 
           (int) rnd, (int) ctr, frame, (int) bdi.eot, (int) bdi.eoi, is_valid, is_write);
#endif
}

/* ========================================================================== */

byte_t getFrame(public_bus bdi)
{
#pragma HLS INLINE
    if (bdi.type != TYPE_AD)
        return FRAMEBITS11;
    else if (!bdi.eot)
        return FRAMEBITS00;
    else
        return FRAMEBITS01;
}

#define ROL(a, offset) (tKeccakLane)((offset != 0) ? ((((tKeccakLane)a) << offset) ^ (((tKeccakLane)a) >> (sizeof(tKeccakLane)*8-offset))) : a)


void theta(tKeccakLane *A)
{
#pragma HLS INLINE    
    unsigned int x, y;
    tKeccakLane C[5], D[5];
#pragma HLS ARRAY_RESHAPE variable=C complete dim=1    
#pragma HLS ARRAY_RESHAPE variable=D complete dim=1    

    for(x=0; x<5; x++) {
    #pragma HLS UNROLL
        C[x] = 0; 
        for(y=0; y<5; y++) {
        #pragma HLS UNROLL
            //printf(" - A[index(%d,%d)] - A[%d] - %02X\n", x, y, index(x,y), (int)A[index(x,y)]);
            C[x] ^= A[index(x, y)];
        }
        //printf("C[%d] - %02X\n", x, (int)C[x]);
    }
    for(x=0; x<5; x++) {
    #pragma HLS UNROLL
        D[x] = ROL(C[(x+1)%5], 1) ^ C[(x+4)%5];
        //printf("D[%d] - %02X\n", x, (int)D[x]);
    }
    for(x=0; x<5; x++) {
    #pragma HLS UNROLL
        for(y=0; y<5; y++) {
        #pragma HLS UNROLL
            A[index(x, y)] ^= D[x];
            //printf("A[%d,%d] - %04X\n", x, y, (int)A[index(x,y)]);
        }
    }
}

void rho(tKeccakLane *A)
{
#pragma HLS INLINE    
    int x, y;
    
    const ap_uint<4> KeccakRhoOffsets[25] = {
        0,1,14,12,
        11,4,12,6,
        7,4,3,10,
        11,9,7,9,
        13,15,5,8,
        2,2,13,8,
        14
    };

    for(x=0; x<5; x++) {
    #pragma HLS UNROLL
        for(y=0; y<5; y++) {
        #pragma HLS UNROLL
            //printf("%d - %d\n", (int) index(x,y), (int) KeccakRhoOffsets[index(x,y)]);
            A[index(x, y)] = ROL(A[index(x, y)], KeccakRhoOffsets[index(x, y)]);
        }
    }
}


void pi(tKeccakLane *A)
{
#pragma HLS INLINE    
    unsigned int x, y;
    tKeccakLane tempA[25];
#pragma HLS ARRAY_RESHAPE variable=tempA complete dim=1    

    for(x=0; x<5; x++) {
	#pragma HLS UNROLL
    	for(y=0; y<5; y++) {
		#pragma HLS UNROLL
			tempA[index(x, y)] = A[index(x, y)];
		}
    }
    for(x=0; x<5; x++) {
	#pragma HLS UNROLL
    	for(y=0; y<5; y++) {
		#pragma HLS UNROLL
			A[index(0*x+1*y, 2*x+3*y)] = tempA[index(x, y)];
    	}
    }
}

void chi(tKeccakLane *A)
{
#pragma HLS INLINE    
    unsigned int x, y;
    tKeccakLane C[5];
#pragma HLS ARRAY_RESHAPE variable=C complete dim=1    

    for(y=0; y<5; y++) { 
    #pragma HLS UNROLL
        for(x=0; x<5; x++) {
        #pragma HLS UNROLL        
            C[x] = A[index(x, y)] ^ ((~A[index(x+1, y)]) & A[index(x+2, y)]);
        }
        for(x=0; x<5; x++) {
        #pragma HLS UNROLL
            A[index(x, y)] = C[x];
        }
    }
}

void iota(tKeccakLane *A, unsigned int indexRound)
{
#pragma HLS INLINE    
    const tKeccakLane KeccakRoundConstants[32] = {
        0x0001, 0x8082, 0x808A, 0x8000, 
        0x808B, 0x0001, 0x8081, 0x8009,
        0x008A, 0x0088, 0x8009, 0x000A,
        0x808B, 0x008B, 0x8089, 0x8003,
        0x8002, 0x0080, 0x800A, 0x000A,
        0x8081, 0x8080, 0x0001, 0x8008,
        0x8082, 0x800A, 0x0003, 0x0009,
        0x8082, 0x8009, 0x0080, 0x8083};
#pragma HLS RESOURCE variable=KeccakRoundConstants core=ROM_1P_1S        
    A[index(0, 0)] ^= KeccakRoundConstants[indexRound];
}

void inversePi(tKeccakLane *B, tKeccakLane *A)
{
#pragma HLS INLINE
    for (int X=0; X<5; X++) {
    #pragma HLS UNROLL
        for (int Y=0; Y<5; Y++) {
        #pragma HLS UNROLL
            int x = (1*X + 3*Y)%5;
            int y = (1*X + 0*Y)%5;
            B[index(x,y)] = A[index(X,Y)];
        }
    }
}

void inversePi(tKeccakLane *A)
{
#pragma HLS INLINE
    tKeccakLane B[25];
    for (int X=0; X<5; X++) {
    #pragma HLS UNROLL
        for (int Y=0; Y<5; Y++) {
        #pragma HLS UNROLL
            int x = (1*X + 3*Y)%5;
            int y = (1*X + 0*Y)%5;
            B[index(x,y)] = A[index(X,Y)];
        }
    }
    for (int X=0; X<25; X++) {
    #pragma HLS UNROLL
        A[X] = B[X];
    }
}

void directPi(tKeccakLane *B, tKeccakLane *A)
{
#pragma HLS INLINE
    for (int X=0; X<5; X++) {
    #pragma HLS UNROLL
        for (int Y=0; Y<5; Y++) {
        #pragma HLS UNROLL
            int x = (0*X + 1*Y)%5;
            int y = (2*X + 3*Y)%5;
            B[index(x,y)] = A[index(X,Y)];
        }
    }
}


void KeccakRound(tKeccakLane *state, unsigned int indexRound)
{
#pragma HLS INLINE
#if DBG_ROUND
    ppp("bef   theta: ", state);
#endif
    theta(state);
#if DBG_ROUND
    ppp("after theta: ", state);
#endif
    rho(state);
#if DBG_ROUND
    ppp("after rho  : ", state);
#endif
    pi(state);
#if DBG_ROUND
    ppp("after pi   : ", state);
#endif
    chi(state);
#if DBG_ROUND
    ppp("after chi  : ", state);
#endif
    iota(state, indexRound);
#if DBG_ROUND
    ppp("after iota : ", state);
#endif
}


/* ------------------------------------------------------------------------- */

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    volatile bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE port=msg_auth_valid axis
#if __SYNTHESIS__
#else
    size_t dbg_loop = 0;
#endif

    /* Interface related temporary variables */
    public_bus bdi, bdi_r;
    output_bus bdo;
    secret_bus key;
    static bool decrypt_r;

    static tKeccakLane key_reg[8];
    static tKeccakLane state[nrLanes];
    ap_uint<5> rnd = 0;
    bool       is_write = 0;

#pragma HLS ARRAY_RESHAPE variable=key_reg complete dim=1    
#pragma HLS ARRAY_RESHAPE variable=state complete dim=1    
    
    fsms_t fsm = S_START;
    ap_uint<3> ctr = 0;
    ap_uint<8> frame = 0;
    bool is_valid = 1;
    decrypt_r = decrypt;

    // Key initialization (if needed)
    if (key_update == 1) {
#if DBG
    printf("Updating key ...\n");
#endif
        LOAD_KEY: for (size_t i=0; i<4; i++) {
            key_bus.read(key);
            for (size_t j=0; j<2; j++) {
            #pragma HLS UNROLL
                key_reg[i*2+j] = (tKeccakLane) (key.data >> (1-j)*LANE_SIZE);
            }
        }
    }
    
    // KeyPack
    for(size_t i=0; i<9;i++) {
    #pragma HLS UNROLL
        if (i == 0) {
          state[i] = (key_reg[i] & 0xFF00) | 0x12;
        } else if (i < 8) {
          state[i] = (key_reg[i] & 0xFF00 ) | (key_reg[i-1] & 0xFF);
        } else {
          state[i] = 0x0100  | (key_reg[i-1] & 0xFF);
        }
    }
    
    for(size_t j=0; j<8; j++) {
        bdi_bus.read(bdi); 
        pp("reading ...\n");
        tKeccakLane hi = (tKeccakLane) (bdi.data >> 16);
        tKeccakLane lo = (tKeccakLane) (bdi.data);
        state[j*2+9+0] = (hi >> 8) | (hi << 8);
        if (j < 7) {
          state[j*2+9+1] = (lo >> 8) | (lo << 8);
        } else {
          state[j*2+9+1] = (lo >> 8) | 0x8100;
        }
    }
    bdi_r = bdi;
    inversePi(state);
#if DBG
    ppp("Init: ", state);
#endif
    
    rnd = Ket_StartRoundOffset;
    bool loop = 1;
    MAINLOOP: while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 50) return;
        else dbg_loop++;
#endif
        /* Default values */
        tKeccakLane s[nrLanes];
#pragma HLS ARRAY_RESHAPE variable=s complete dim=1    

        // Copy state
        for(size_t i=0; i<25; i++) {
            #pragma HLS UNROLL
            s[i] = state[i];
        }

        #if DBG
        printHeader(fsm, (int)rnd, (int)ctr, (int)frame, bdi_r, is_valid, is_write);
        #endif
        // Input
        data_t input = (fsm == S_TAG) 
                        ? (data_t) 0 
                        : (data_t) (((bdi_r.data & 0x000000FF) << 24) |
                                    ((bdi_r.data & 0x0000FF00) <<  8) |
                                    ((bdi_r.data & 0x00FF0000) >>  8) |
                                    ((bdi_r.data & 0xFF000000) >> 24));

        data_t padLoc;
        if (fsm == S_TAG)
            padLoc = (data_t) frame;
        else
            switch(bdi_r.size) {
            case 0:  padLoc = (data_t) frame; break;
            case 1:  padLoc = ((data_t) frame << 8); break;
            case 2:  padLoc = ((data_t) frame << 16); break;
            case 3:  padLoc = ((data_t) frame << 24); break;
            default: padLoc = 0;
            }

        byte_t padding = ((bdi_r.size == 4) && (fsm != S_TAG)) 
                         ? (byte_t) (0x08 ^ frame) : (byte_t) 0x08;
        
        tKeccakLane padLoc_hi = padLoc & 0xFFFF;
        tKeccakLane padLoc_lo = padLoc >> 16;

        tKeccakLane input_hi = input & 0xFFFF;
        tKeccakLane input_lo = input >> 16;

        byte_t byte0 = (is_write && decrypt_r && ((bdi_r.valid_bytes >> 3) & 0x1) ) 
                        ? (byte_t) 0 : (byte_t) s[0];
        byte_t byte1 = (is_write && decrypt_r && ((bdi_r.valid_bytes >> 2) & 0x1) ) 
                        ? (byte_t) 0 : (byte_t) (s[0] >> 8);
        byte_t byte2 = (is_write && decrypt_r && ((bdi_r.valid_bytes >> 1) & 0x1) ) 
                        ? (byte_t) 0 : (byte_t) s[6];
        byte_t byte3 = (is_write && decrypt_r && ((bdi_r.valid_bytes >> 0) & 0x1) ) 
                        ? (byte_t) 0 : (byte_t) (s[6] >> 8);
        tKeccakLane state_hi = (tKeccakLane) byte0 | ((tKeccakLane) byte1 << 8);
        tKeccakLane state_lo = (tKeccakLane) byte2 | ((tKeccakLane) byte3 << 8);

        tKeccakLane enc_hi = state[0] ^ input_hi ^ padLoc_hi;
        tKeccakLane enc_lo = state[6] ^ input_lo ^ padLoc_lo;
        tKeccakLane in_hi = state_hi ^ input_hi ^ padLoc_hi;
        tKeccakLane in_lo = state_lo ^ input_lo ^ padLoc_lo;

        if ((S_DATA == fsm) 
            || (S_STRIDE == fsm && rnd == Ket_StrideRoundOffset)
            || (S_TAG == fsm))  {
            s[0] = in_hi;
            s[6] = in_lo;
            s[12] ^= (tKeccakLane) padding;
        } else {
            s[0] ^= (tKeccakLane) frame;
        }

        /* ==================== */
        /* Preparing input data */
        /* ==================== */
        #if DBG
        ppp("Input  : ", s);
        #endif
        
        /* Main operation */
        KeccakRound(s, rnd);


        #if DBG
        if (rnd == maxRound)
           ppp("Output : ", s);
        #endif
        
        /* ========= */
        /* Post processing  */
        /* ========= */
        tKeccakLane out_hi = (is_write) ? enc_hi : state[0];
        tKeccakLane out_lo = (is_write) ? enc_lo : state[6];

        data_t output = (((data_t)  (out_hi & 0x00FF)) << 24) |
                         (((data_t) (out_hi & 0xFF00)) <<  8) |
                         (((data_t) (out_lo & 0x00FF)) <<  8) |
                         (((data_t) (out_lo & 0xFF00)) >>  8);



        // Output write
        bool is_valid_next = 0;
        if (is_write || (fsm == S_TAG)) {
            if (S_TAG == fsm && decrypt_r) { // Standard output 
                bool valid = (output == bdi_r.data) ? 1:0;
                is_valid_next = is_valid & valid;
            } else {
                bdo.data = output;
                bdo_bus.write(bdo);
                pp("writing...\n");
            }
        }

        bool next_inverse = false;
        fsms_t next_fsm = fsm;
        byte_t next_frame = 0;
        bool   next_write = 0;
        switch (fsm) {
        case S_START:
            if (maxRound == rnd) {
                next_fsm = S_DATA;
                bdi_bus.read(bdi);
                pp("reading ...\n");
                next_frame = getFrame(bdi);
                bdi_r = bdi;
                rnd = maxRound;
            } else {
                next_inverse = false;
                rnd++;
            }
            break;
        case S_DATA:
            bdi_bus.read(bdi);
            pp("reading ...\n");
            if (bdi.type == TYPE_D && bdi.size != 0)
                next_write = 1;
            if (bdi.type == TYPE_D && bdi.eot) {
                next_fsm = S_STRIDE;
                next_frame = FRAMEBITS10;
                rnd = Ket_StrideRoundOffset;
            } else {
                next_frame = getFrame(bdi);
            }                        
            bdi_r = bdi;
            break;
        case S_STRIDE:
            if (maxRound == rnd) {
                next_fsm = S_TAG;
                next_frame = FRAMEBITS0;
                if (decrypt_r) {
                    bdi_bus.read(bdi);
                    pp("reading ...\n");
                    bdi_r = bdi;
                }
            } else {
                next_inverse = false;
                rnd++;
            }
            break;            
        default: // S_TAG
            next_fsm = fsm;
            next_frame = FRAMEBITS0;
            if (ctr < 3 && decrypt_r) {
                bdi_bus.read(bdi);
                pp("reading ...\n");
                bdi_r = bdi;
            }
            if (ctr == 3 && decrypt_r) *msg_auth_valid = is_valid_next;
            is_valid = is_valid_next;
            if (ctr == 3) loop = 0;
            else ctr ++;            
        }

        for(size_t i=0; i<25; i++) {
            #pragma HLS UNROLL
            state[i] = s[i];
        }

        fsm = next_fsm;
        frame = next_frame;
        is_write = next_write;
    }
    return;
}
